/**
***	$RCSfile: dateops.c,v $
***	$Revision: 1.18 $
***	$Date: 2011/08/04 12:49:06 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/

/**
*** add "nthday" to give "dayname nth-day-of-month suffix"
*** ex: "Monday 1 st"; should accept yyyymmdd like other
*** command
**/

/**
*** in bmktime, consider changing non-convergence algorithm to use
*** a counter; this avoids problems with word size and overflow
**/


/**
*** Uncomment is needed:
***
*** #define OS_SOLARIS	1
***
*** #define OS_IRIX	1
**/


#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>


static char RCSid[] =
    "$Id: dateops.c,v 1.18 2011/08/04 12:49:06 adamm Exp adamm $";


void adjust(char *, char *, char *);
void bmktime(char *, char *);
void manpage(void);
void nextmo(char *, char *);
void nextprevdow(char *, char *, char *, char *);
void prevmo(char *, char *);
void unhexify(unsigned char **);

char *Prog;


int main(int argc, char *argv[])
{
	char *arg;
	char buf[64];
	char *dow;
	int max;
	int op;

	/* Make the program name pretty */
	Prog = strrchr(argv[0], '/');
	if (Prog == NULL) {
		Prog = &argv[0][0];
	} else {
		++Prog;
	}

	/* Deal with --help and --man */
	if (argc > 1) {
	    if (
		(strcmp(argv[1], "--help") == 0)
		||
		(strcmp(argv[1], "--man") == 0)
	    ) {
		manpage();
		exit(0);
	    }
	}

	/* Set locale and time zone stuff */
	(void) setlocale(LC_TIME, "");

	/* Do a quick sanity check on the command line */
	if ((argc == 1) || (argc > 4)) {
		(void) fprintf(stderr,
		    "%s: incorrect number of args\n\n", Prog);
		(void) fprintf(stderr,
		    "Use \"%s --man for more details.\n", Prog);
		exit(1);
	}

	/* See what operation we were asked to perform */
	if (strcmp(argv[1], "lastmonth") == 0) {
		op = 0;
	} else if (strcmp(argv[1], "last") == 0) {
		op = 1;		/* same as "next" */
	} else if (strcmp(argv[1], "maketime") == 0) {
		op = 2;
	} else if (strcmp(argv[1], "next") == 0) {
		op = 1;		/* same as "last" */
	} else if (strcmp(argv[1], "nextmonth") == 0) {
		op = 3;
	} else if (strcmp(argv[1], "tomorrow") == 0) {
		op = 4;
	} else if (strcmp(argv[1], "yesterday") == 0) {
		op = 5;
	} else if (
		((argv[1][0] == '-') || (argv[1][0] == '+'))
		&& isdigit(argv[1][1])
	) {
		op = 6;
	} else {
		(void) fprintf(stderr,
		    "%s: unknown operation: %s\n\n", Prog, argv[1]);
		(void) fprintf(stderr,
		    "Use \"%s --man for more details.\n", Prog);
		exit(1);
	}

	/* Another sanity check */
	if ((op != 1) && (argc > 3)) {
		(void) fprintf(stderr,
		    "%s: incorrect number of args\n", Prog);
		(void) fprintf(stderr,
		    "Use \"%s --man for more details.\n", Prog);
		exit(1);
	}

	/* Figure out what to pass as an argument */
	if (op == 1) {
		dow = argv[2];
		if (argc == 3) {
			arg = NULL;
		} else {
			arg = argv[3];
		}
	} else {
		if (argc == 2) {
			arg = NULL;
		} else {
			arg = argv[2];
		}
	}

	/* Finally, perform the requested operation */
	switch (op) {
	case 0:
		prevmo(arg, buf);
		break;

	case 1:
		if (strcmp(argv[1], "last") == 0) {
			nextprevdow("-1", dow, arg, buf);
		} else {
			nextprevdow("+1", dow, arg, buf);
		}
		break;

	case 2:
		bmktime(arg, buf);
		break;

	case 3:
		nextmo(arg, buf);
		break;

	case 4:
		adjust("+1", arg, buf);
		break;

	case 5:
		adjust("-1", arg, buf);
		break;

	case 6:
		adjust(argv[1], arg, buf);
		break;
	}

	(void) printf("%s\n", buf);
	exit(0);
} /* main */



/**
*** In adjust(), why not just add or subtract 86,400 (one day's worth of
*** seconds and be done with it? If the program runs at the wrong time
*** on the day we change to/from Daylight Savings Time, it's possible
*** we'll advance or retreat two days instead of one. Moving by
*** increments of 11 hours guarantees we always get the correct answer
*** and optimizes over doing it one hour at a time.
**/

void adjust(char *adj, char *arg, char *out)
{
	char buf[16];
	int curr;
	int ndays;
	time_t now;
	int sign;
	struct tm *tm;
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Break the adjustment into days and sign (direction) */
	ndays = atoi(adj);
	if (ndays < 0) {
		sign = -1;
		ndays *= sign;
	} else {
		sign = 1;
	}

	/* Pick a starting point */
	if (arg == NULL) {
		if (time(&now) == -1) {
			perror(Prog);
		}

		tm = localtime(&now);
		curr = tm->tm_mday;
	} else {
		if (strlen(arg) < 8) {
			(void) strcpy(out, "bad date");
			return;
		}

		bmktime(arg, buf);

		/* If sizeof(time_t) > sizeof(unsigned long), we're screwed! */
		now = (time_t) strtoul(buf, (char **)NULL, 10);
		tm = localtime(&now);
		curr = tm->tm_mday;
	}

	while (ndays > 0) {
		do {
			now += (sign * 39600); /* 8 hours */
			tm = localtime(&now);
		} while (tm->tm_mday == curr);

		curr = tm->tm_mday;
		--ndays;
	}

	(void) strftime(out, 16, "%Y%m%d", tm);
	return;
} /* adjust */


void bmktime(char *arg, char *out)
{
	char buf[16];
	char fmt[] = "%Y%m%d%H%M%S";
	time_t half;
	int i;
	time_t max;
	time_t min;
	int n;
	time_t now;
	char target[16];
	struct tm *tm;
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Today */
	if (time(&now) == -1) {
		perror(Prog);
	}

	/* Quick and dirty */
	if (arg == NULL) {
		(void) sprintf(out, "%lu", now);
		return;
	}

	/* Look more closely at the argument we got */
	if (strlen(arg) < 8) {
		exit(1);
	}

	/* Finish processing today */
	tm = localtime(&now);
	(void) strftime(buf, sizeof(buf), fmt, tm);

	/* The target date */
	(void) memset(target, 0, 16);
	(void) strcpy(target, arg);
	n = strlen(target);
	if (n < 14) {
		i = 14 - n;
		(void) memset(&target[n], '0', i);
	}

	/* Set min and max */
	if (strcmp(target, buf) > 0) {
		/* target is in the future, today = min */
		min = now;
		max = (time_t) 2147483647;
	} else {
		/* target is in the past, today = max */
		min = (time_t) 0;
		max = now;
	}

	/* Converge on the target using a binary search */
	for (;;) {
		/* Can't do "(min + max) / 2" because ofoverflow */
		half = (max - min) / 2;
		half += min;
		tm = localtime(&half);
		(void) strftime(buf, sizeof(buf), fmt, tm);

		n = strcmp(target, buf);
		if (n == 0) {
			break;
		} else if (n < 0) {
			max = half - 1;
		} else {
			min = half + 1;
		}
		if (min > max) {
		    (void) fprintf(
			stderr, "%s: unable to converge\n", Prog
		    );
		    exit(1);
		}
	}

	(void) sprintf(out, "%lu", half);
	return;
} /* bmktime */


void manpage(void)
{
	unsigned char *lines[] = {
	    #include "dateops.man.hex"
	    0
	};

	unhexify(lines);
} /* manpage */


void nextmo(char *arg, char *out)
{
	char buf[16];
	int curr;
	time_t incr;
	time_t now;
	struct tm *tm;
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Pick a starting point */
	if (arg == NULL) {
		if (time(&now) == -1) {
			perror(Prog);
		}

		tm = localtime(&now);
		curr = tm->tm_mon;
	} else {
		if (strlen(arg) < 8) {
			(void) strcat(arg, "01");
		}

		bmktime(arg, buf);

		/* If sizeof(time_t) > sizeof(unsigned long), we're screwed! */
		now = (time_t) strtoul(buf, (char **)NULL, 10);
		tm = localtime(&now);
		curr = tm->tm_mon;
	}

	incr = (time_t) (7 * 86400); /* 1 week */
	do {
		now += incr;
		tm = localtime(&now);
	} while (tm->tm_mon == curr);

	(void) strftime(out, 16, "%Y%m", tm);
	return;
} /* nextmo */


void nextprevdow(char *n, char *dow, char *arg, char *out)
{
	char buf[16];
	int day;
	time_t now;
	struct tm *tm;
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Convert the day name to a number */
	if (strncasecmp(dow, "sun", 3) == 0) {
		day = 0;
	} else if (strncasecmp(dow, "mon", 3) == 0) {
		day = 1;
	} else if (strncasecmp(dow, "tue", 3) == 0) {
		day = 2;
	} else if (strncasecmp(dow, "wed", 3) == 0) {
		day = 3;
	} else if (strncasecmp(dow, "thu", 3) == 0) {
		day = 4;
	} else if (strncasecmp(dow, "fri", 3) == 0) {
		day = 5;
	} else if (strncasecmp(dow, "sat", 3) == 0) {
		day = 6;
	} else {
		(void) strcpy(out, "bad day name");
		return;
	}

	/* Pick a starting point */
	if (arg == NULL) {
		if (time(&now) == -1) {
			perror(Prog);
		}

		tm = localtime(&now);
		(void) strftime(out, 16, "%Y%m%d", tm);
	} else {
		if (strlen(arg) < 8) {
			(void) strcpy(out, "bad date");
			return;
		}

		(void) strcpy(out, arg);
	}

	do {
		adjust(n, out, buf);
		bmktime(buf, out);
		now = (time_t) strtoul(out, (char **)NULL, 10);
		tm = localtime(&now);
		(void) strftime(buf, 16, "%w", tm);
		(void) strftime(out, 16, "%Y%m%d", tm);
	} while (atoi(buf) != day);

	(void) strftime(out, 16, "%Y%m%d", tm);
	return;
} /* nextprevdow */


void prevmo(char *arg, char *out)
{
	char buf[16];
	int curr;
	time_t incr;
	time_t now;
	struct tm *tm;
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Pick a starting point */
	if (arg == NULL) {
		if (time(&now) == -1) {
			perror(Prog);
		}

		tm = localtime(&now);
		curr = tm->tm_mon;
	} else {
		if (strlen(arg) < 8) {
			(void) strcat(arg, "01");
		}

		bmktime(arg, buf);

		/* If sizeof(time_t) > sizeof(unsigned long), we're screwed! */
		now = (time_t) strtoul(buf, (char **)NULL, 10);
		tm = localtime(&now);
		curr = tm->tm_mon;
	}

	incr = (time_t) (7 * 86400); /* 1 week */
	do {
		now -= incr;
		tm = localtime(&now);
	} while (tm->tm_mon == curr);

	(void) strftime(out, 16, "%Y%m", tm);
	return;
} /* prevmo */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            (void) fputc((int) c, stdout);
        }
    }
} /* unhexify */


/*	===== MAN PAGE STARTS HERE =====
NAME
    dateops -- perform simple date operations

SYNOPSIS
    dateops op [ date ]

DESCRIPTION
    dateops will perform simple date operations (listed below). If no
    date is given the operation will be performed relative to today;
    otherwise, the date given will be used.

    The date is as much of YYYYMM [ DD [ hh [ mm [ ss ] ] ] ] as makes
    sense.

OPERATORS
    maketime
	converts the date (and time, if given) to "epoch seconds". Any
	portion of the time not given will be treated as "00"; if no
	date is given, the current date and time will be used.

    +#, -#
	adds or subtracts # days; gives YYYYMMDD

    tomorrow
	equivalent to "+1"

    yesterday
	equivalent to "-1"

    nextmonth
	gives YYYYMM of the following month

    lastmonth
	gives YYYYMM of the preceeding month

EXIT STATUS
    dateops will exit with one of the following values:

	0    no errors
	1    incorrect number of arguments

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====	*/


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
