/**
***	$RCSfile: edate.c,v $
***	$Revision: 1.14 $
***	$Date: 2010/01/09 21:29:55 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


/**
*** Uncomment as needed:
***
*** #define OS_SOLARIS	1
***
*** #define OS_IRIX	1
**/


#include <sys/param.h>
#include <sys/time.h>
#include <ctype.h>
#include <locale.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>


#if defined(OS_SOLARIS)
    extern char *optarg;
    extern int optind, opterr, optopt;
#endif


static char RCSid[] =
    "$Id: edate.c,v 1.14 2010/01/09 21:29:55 adamm Exp adamm $";


void manpage(void);
void suffix(int, char *);
void unhexify(unsigned char **);

char *Prog;


int main(int argc, char *argv[])
{
	int c;
	char buf[8192];
	int f;
	char fmt[8192];
	int got_t = 0;
	int h;
	struct tm *herenow;
	char hold[8192];
	int n;
	struct timeval now;
	char *p;
	char sfx[4];
#if defined(OS_IRIX) || defined(OS_SOLARIS)
	char tzenv[16];
#endif

	/* Make the program name pretty */
	Prog = strrchr(argv[0], '/');
	if (Prog == NULL) {
		Prog = &argv[0][0];
	} else {
		++Prog;
	}

	/* Deal with --help and --man */
	if (argc > 1) {
	    if (
		(strcmp(argv[1], "--help") == 0)
		||
		(strcmp(argv[1], "--man") == 0)
	    ) {
		manpage();
		exit(0);
	    }
	}

	(void) setlocale(LC_TIME, "");
	(void) strcpy(fmt, "%a %b %e %T %Z %Y");

	/* Deal with options */
	while ((c = getopt(argc, argv, ":t:u")) != -1) {
		switch ((char)c) {
		case 't':
			/* ought to be strtoumax()! */
			/* ought to do better error checking */
			now.tv_sec =
			    (time_t) strtol(optarg, (char **)NULL, 10);
			now.tv_usec = (suseconds_t) 0;
			got_t = 1;
			break;

		case 'u':
#if defined(OS_IRIX) || defined(OS_SOLARIS)
			(void) strcpy(tzenv, "TZ=UTC0");
			/* Ought to check the return value. Someday. */
			(void) putenv(tzenv);
#else
			(void) setenv("TZ", "UTC0", 1);
#endif
			break;

		default:
			(void) fprintf(stderr,
			    "%s: unrecognized switch: '%c'\n\n",
			    Prog, (char) optopt);
			(void) fprintf(stderr,
			    "Use \"%s --man\" for more details.\n",
			    Prog);
			exit(1);
		}
	}

	/* Is there more than one arg left on the command line? */
	if ((argc - optind) > 1) {
		(void) fprintf(stderr,
		    "%s: incorrect number of args\n\n", Prog);
		(void) fprintf(stderr,
		    "Use \"%s --man\" for more details.\n", Prog);
		exit(1);
	}


	/* If there's exactly one arg left . . . */
	if (argc > optind) {

	    /* and it's not an empty string . . . */
	    if (*argv[optind] != '\0') {

		/* and it starts with "+" . . .*/
		if (*argv[optind] != '+') {
			(void) fprintf(stderr,
			    "%s: invalid format: '%s'\n",
			    Prog, argv[optind]);
			(void) fprintf(stderr,
			    "%s: format must start with '+'\n\n",
			    Prog, argv[optind]);
			(void) fprintf(stderr,
			    "Use \"%s --man\" for more details.\n", Prog);
			exit(1);
		}

		/* it must be our format string -- so save it */
		(void) strcpy(fmt, &argv[optind][1]);
	    }
	}

	/* Only get the current time if we didn't get -t */
	if (got_t == 0) {
	    (void) gettimeofday(&now, NULL);
	}

	/* Yes, even if we got -u -- otherwise we'd print the */
	/* wrong timezone                                     */
	herenow = localtime(&now.tv_sec);

	/* Copy fmt[] to buf[] then copy it back one chararacter  */
	/* at a time, making substitutions for %E, %f, and %s as  */
	/* we go                                                  */

	(void) strcpy(hold, fmt);
	n = strlen(hold);

	f = 0;
	fmt[f] = '\0';

	for (h = 0; h < n; h++) {
	    if (hold[h] != '%') {
		fmt[f++] = hold[h];
		continue;
	    }

	    switch (hold[h + 1]) {
	    case 'E':
		(void) sprintf(buf, "%d", herenow->tm_mday);
		break;

	    case 'f':
		(void) suffix(herenow->tm_mday, sfx);
		(void) sprintf(buf, "%s", sfx);
		break;

	    case 's':
		(void) sprintf(buf, "%lu", now.tv_sec);
		break;

	    case 'U':
		(void) sprintf(buf, "%lu", now.tv_usec);
		break;

	    default:
		fmt[f++] = hold[h];
		continue;
	    }

	    fmt[f] = '\0';
	    (void) strcat(fmt, buf);
	    f = strlen(fmt);
	    ++h;
	}

	/* Now that we've removed all non-standard stuff from */
	/* fmt[], let strftime() to the heavy lifting         */
	fmt[f] = '\0';
	(void) strftime(buf, sizeof(buf), fmt, herenow);
	(void) printf("%s", buf);

	/* Final \n */
	(void) printf("\n");

	exit(0);
} /* main */


void manpage(void)
{
	unsigned char *lines[] = {
	    #include "edate.man.hex"
	    0
	};

	unhexify(lines);
} /* manpage */


void suffix(int d, char *buf)
{
    switch (d) {
    case  1:
    case 21:
    case 31:
	strcpy(buf, "st");
	break;

    case  2:
    case 22:
	strcpy(buf, "nd");
	break;

    case  3:
    case 23:
	strcpy(buf, "rd");
	break;

    default:
	strcpy(buf, "th");
	break;
    }
} /* suffix */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }
} /* unhexify */


/*	===== MAN PAGE STARTS HERE =====
NAME
    edate -- POSIX date(1), with extensions (and deletions)

SYNOPSIS
    edate [ -u ] [ -t seconds ] [ +format ]

DESCRIPTION
    edate emulates the date(1) command, as specified by POSIX (formally,
    The Open Group Base Specifications Issue 6 IEEE Std 1003.1, 2004
    Edition), with the extensions described below.

    edate does *not* support the X/Open System Interfaces Extension
    functionality of setting the system time.

OPTIONS
    -t seconds
	instead of formatting and printing the current date and time,
	use the date and time specified; the units are "epoch seconds"
	("the UNIX way").

CONVERSION SPECIFICATIONS
    %E
	replaced by the day of the month as a decimal number [1,31];
	a single digit is *not* preceded by a space. [tm_mday]

    %f
	replaced by (English) suffix appropriate for the day of the
	month ["st", "nd", "rd", "th"].
 
    %s (epoch seconds)
	replaced by "epoch seconds" as a decimal number.

    %U
	replaced by microseconds; if used with -t, always replaced
	with zero.

EXIT STATUS
    (real soon now)

SEE ALSO
    date(1)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====	*/


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
