/**
***	$RCSfile: hexify.c,v $
***	$Revision: 1.7 $
***	$Date: 2009/12/11 12:59:19 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

static char RCSid[] =
    "$Id: hexify.c,v 1.7 2009/12/11 12:59:19 adamm Exp adamm $";

/* Keep/remove leading colon as desired, or s/:/?/ */
static char OPTS[] = ":f";
char	*Prog;

void	encodeFile	(char *);
void	showFunction	(void);
void	usage		(FILE *);


int main(int argc, char *argv[])
{
    int  c;

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	    Prog = &argv[0][0];
    } else {
	    ++Prog;
    }

    /* Deal with --help and --man */
    if (argc > 1) {
	if (
	    (strcmp(argv[1], "--help") == 0)
	    ||
	    (strcmp(argv[1], "--man") == 0)
	) {
	    usage(stdout);
	    exit(0);
	}
    }

    /* Parse the command line */
    if (argc > 2) {
	usage(stderr);
	exit(1);
    }

    while ((c = getopt(argc, argv, OPTS)) != -1) {
	switch (c) {
	case 'f':
	    showFunction();
	    exit(0);

	default:
	    (void) fprintf(stderr,
		"%s: unrecognized switch: '%c'\n\n", Prog, (char) optopt);
	    usage(stderr);
	    exit(1);
	}
    }

    /* Did we get a file? */
    if (argc == 1) {
	encodeFile("");
    } else {
	encodeFile(argv[optind]);
    }

    exit(0);
} /* main */


void encodeFile(char *file)
{
    int			b;
    ssize_t		bytes;
    unsigned char	buf[36];
    int			fd;

    /* (Try to) Open the file */
    if (strlen(file) == 0) {
	fd = 0;
    } else {
	fd = open(file, O_RDONLY);
	if (fd == -1) {
	    fprintf(stderr, "%s: %s: %s\n",
		Prog, strerror(errno), file);
	    exit(2);
	}
    }

    /* Read the file one "line" at a time */
    for (;;) {
	bytes = read(fd, buf, (ssize_t) 36);
	if (bytes == (ssize_t) 0) {
	    break;
	}

	printf("    \"");
	for (b = 0; b < bytes; b++) {
	    printf("%02x", buf[b]);
	}
	printf("\",\n");
    }

    /* Clean up and quit */
    (void) close(fd);
    return;
} /* encodeFile */


void showFunction()
{
    (void) printf(
"\n"
"\n"
"void unhexify(unsigned char **);\n"
"\n"
"\n"
"unsigned char *lines[] = {\n"
"    #include \"your-file-here\"\n"
"    0\n"
"};\n"
"\n"
"\n"
"unhexify(lines);\n"
"\n"
"\n"
"void unhexify(unsigned char **ptr)\n"
"{\n"
"    int hex[] = {\n"
"        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,\n"
"        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,\n"
"        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,\n"
"        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,\n"
"        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,\n"
"        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,\n"
"        0, 10, 11, 12, 13, 14, 15\n"
"    };\n"
"\n"
"    int b;\n"
"    int c;\n"
"\n"
"    for ( ; *ptr; *(ptr++)) {\n"
"        for (b = 0; b < strlen(*ptr); b += 2) {\n"
"            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];\n"
"            putc((int) c, stdout);\n"
"        }\n"
"    }\n"
"\n"
"    return;\n"
"} /* unhexify */\n"
    );

    return;
} /* showFunction */


void usage(FILE *stream)
{
    (void) fprintf(stream,
	"Usage:\n"
	"    %s [ file ]\n"
	"        encode file (or stdin)\n\n"
	"    %s -f\n"
	"        emit decoding function\n",
	Prog, Prog
    );
    return;
} /* Usage */


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
