/**
***	$RCSfile: isprint.c,v $
***	$Revision: 1.8 $
***	$Date: 2010/01/08 21:41:32 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>


static char RCSid[] =
    "$Id: isprint.c,v 1.8 2010/01/08 21:41:32 adamm Exp adamm $";


void error(char *);
void report(char *, int, char *, size_t, size_t);
int testFile(char *, int);
void manpage(void);
int processArgs(int, char *[]);
void unhexify(unsigned char **);


#define READBUF	    65536


char *Prog;

int Counts = 0;
int Errors = 1;
int Word   = 1;

int printing[] = {
    /*		0  1  2  3  4  5  6  7  8  9  a  b  c  d  e  f    */
    /* 0x0 */	1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1,
    /* 0x1 */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x2 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x3 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x4 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x5 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x6 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /* 0x7 */	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
    /* 0x8 */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0x9 */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xa */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xb */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xc */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xd */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xe */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
    /* 0xf */	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
};


int main(int argc, char *argv[])
{
    int argx;
    int exitval = 0;
    int numfiles;

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	    Prog = &argv[0][0];
    } else {
	    ++Prog;
    }

    /* Parse the command line */
    argx = processArgs(argc, argv);

    numfiles = argc - argx;
    /* Read the file(s) and/or stdin */
    if (argx == argc) {
	exitval = exitval | testFile("", numfiles);
    } else {
	for ( ; argx < argc; argx++) {
	    if (strcmp(argv[argx], "-") == 0) {
		exitval = exitval | testFile("", numfiles);
	    } else {
		exitval = exitval | testFile(argv[argx], numfiles);
	    }
	}
    }

    exit(exitval);
} /* main */


void error(char *errmsg)
{
    if (Errors == 1) {
	(void) fprintf(stderr, "%s", errmsg);
    }
} /* error */


void manpage(void)
{
    unsigned char *lines[] = {
	#include "isprint.man.hex"
	0
    };

    unhexify(lines);
} /* manpage */


int processArgs(int ac, char *av[])
{
    int c;
    char errmsg[BUFSIZ];
    int mutex;

    /* Special handling for --help and --man */
    if (ac > 1) {
	if (
	    (strcmp(av[1], "--help") == 0)
	    ||
	    (strcmp(av[1], "--man") == 0)
	) {
	    manpage();
	    exit(0);
	}
    }

    mutex = 0;
    while ((c = getopt(ac, av, "eqvV")) != -1) {
	switch (c) {
	case 'e':
	    Errors = 0;
	    break;

	case 'q':
	    Counts = 0;
	    Word = 0;
	    ++mutex;
	    break;

	case 'v':
	    Counts = 1;
	    Word = 0;
	    ++mutex;
	    break;

	case 'V':
	    Counts = 1;
	    Word = 1;
	    ++mutex;
	    break;
	}
    }

    if (mutex > 1) {
	(void) sprintf(errmsg, 
	    "%s: only one of '-q', '-v', and '-V' may be given;"
	    " aborting\n",
	    Prog);
	error(errmsg);
	exit(4);
    }

    return optind;
} /* processArgs */


void report(
    char *filename,
    int numfiles,
    char *string,
    size_t text,
    size_t binary
) {
    if ((Word == 0) && (Counts == 0)) {
	return;
    }

    if (strcmp(filename, "") != 0) {
	(void) printf("%s: ", filename);
    } else {
	(void) printf("(stdin): ");
    }

    if (Word == 1) {
	(void) printf("%s", string);
    }

    if (Counts == 1) {
	if (Word == 1) {
	    printf(" ");
	}

	(void) printf("%lu %lu %lu", (text + binary), text, binary);
    }

    (void) printf("\n");
    return;
} /* report */


int testFile(char *filename, int numfiles)
{
    int c;
    unsigned char buf[READBUF];
    int file;
    struct stat	ino;
    size_t n;
    int ret;
    char string[BUFSIZ];
    size_t tally[2];

    /* If this isn't stdin, stat the file */
    if (strcmp(filename, "") != 0) {
	if (stat(filename, &ino) == -1) {
	    report(filename, numfiles, "error", 0, 0);
	    (void) sprintf(string, "%s: %s: \"%s\" (%d)\n",
		    Prog, strerror(errno), filename, errno);
	    error(string);
	    return 4;
	}

	/* Bail out early if it's not a plain file */
	if (! S_ISREG(ino.st_mode)) {
	    report(filename, numfiles, "other", 0, 0);
	    return 2;
	}
    }

    /* If not stdin, open the file */
    if (strcmp(filename, "") != 0) {
	file = open(filename, O_RDONLY);
	if (file == -1) {
	    report(filename, numfiles, "error", 0, 0);
	    (void) sprintf(string, "%s: %s: \"%s\" (%d)\n",
		    Prog, strerror(errno), filename, errno);
	    error(string);
	    return 4;
	}
    } else {
	file = 0;

    }

    /* Read the file and count the bytes */
    tally[0] = tally[1] = (size_t) 0;
    while ((n = read(file, buf, (size_t) READBUF)) > 0) {
	for (c = 0; c < n; c++) {
	    ++tally[printing[buf[(int) c]]];
	}
    }

    (void) close(file);

    if (tally[1] > 0) {
	(void) strcpy(string, "binary");
	ret = 1;
    } else {
	(void) strcpy(string, "text");
	ret = 0;
    }
    report(filename, numfiles, string, tally[0], tally[1]);

    return ret;
} /* testFile */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
	0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
	0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
	0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
	for (b = 0; b < strlen(*ptr); b += 2) {
	    c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
	    putc((int) c, stdout);
	}
    }
} /* unhexify */


/*	===== MAN PAGE STARTS HERE =====
NAME
    isprint - check if a file contains only printable characters

SYNOPSIS
    isprint [ options ] [ file ... ]

DESCRIPTION
    isprint reads a file and looks for non-printable characters (as
    defined by isprint(3)); if no such characters are found, isprint
    consider the file to be "a text file"; otherwise it's "a binary
    file"; directories and special files are considered "other." If the
    file is a symlink, isprint looks at the link target, not the link
    itself.

    By default, isprint displays one line per file, as follows:

	filename: word

    The value of "word" will be one of "text," "binary," "error," or
    "other."

    If the "-v" flag is given, the output will be as follows:

	filename: total printable non-printing

    where "total," "printable," and "non-printing" are integers
    representing the corresponding number of characters. If "-V" is
    given, the output will be:

	filename: word total printable non-printing

    If "-q" is given, output will be supressed, and the only indication
    of "printability" will be the exit status (described below).

OPTIONS
    -e
    	suppress errors

    -q
	quiet: no output, just an exit value; if multiple files are
	specified, the exit value will be the logical OR of the exit
	values for each file; errors are still printed (on stderr)

    -v
	verbose: print total characters, printable characters,
	non-printing characters

    -V
	extra verbose: print "text" or "binary" followed by the same
	output as "-v" (above)

CAVEATS
    The options "-q", "-s", "-v", and "-V" are mutually exclusive.

EXIT STATUS
    isprint exits with 0 if all the characters in a file are printable,
    with 1 if there are non-printable characters, 2 if the file is not a
    plain file, and 4 if there are any errors. If multiple files are
    given, the exit status will be the bitwise OR of the exit status of
    each file.

SEE ALSO
    file(1), isprint(3)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====	*/


/**
*** Copyright, 2010, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
