/**
***	$RCSfile: lstat.c,v $
***	$Revision: 1.71 $
***	$Date: 2012/01/05 03:05:28 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


/**
*** to-do:
*** - fix CPP stuff
***	    (um, what CPP stuff?)
*** - fill in lstat.h (mostly done)
*** - handle time variants
*** - show man page
**/


#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <grp.h>
#include <limits.h>
#include <pwd.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "lstat.h"

#if (! defined(OS_IRIX)) && (! defined(OS_SOLARIS))
    #include <stdint.h>
#endif


/* Declarations */
void dfltMembers(char *);
void dfltMembersHuman(char *);
void dfltStrs(void);
int formatPermsChmod(
    char *, size_t, mode_t,
    int, int, int, int, char *, char *, int
);
int formatPermsLs(
    char *, size_t, mode_t,
    int, int, int, int, char *
);
void manpage(void);
int processMembers(char *, Member *, int *);
void setBaseCase(char *, int, int *, int *);
void setFormat(char *, int, char *);
void setSeparators(char *, char *, char *, char *, char *);
void showDfltMembers(void);
void showHumanMembers(void);
int showLsMode(char *, size_t, mode_t);
int showMembers(char *, struct stat *, Member *, int);
int showNumeric(char *, size_t, UMAXTYPE__T, Member);
int showSymbolicGid(char *, size_t, void *);		/* FIX */
int showSymbolicPerm(char *, size_t, mode_t);
int showSymbolicType(char *, size_t, mode_t, char);
int showSymbolicTypePerm(char *, size_t, mode_t);
int showSymbolicUid(char *, size_t, void *);		/* FIX */
int showTarget(char *, char *, size_t, mode_t);
void splitQuotes(char *, char *, char *);
char *strnsubstr(char *, const char *, size_t, size_t);
int typeToStr(char *, size_t, void *, char *, Member);
void unhexify(unsigned char **);
void usage(void);
int validateMember(char *, char *, int *, int *);


/* Globals */
char fileQuoteL[SHORT_BUF];
char fileQuoteR[SHORT_BUF];
char inputFormatSep[SHORT_BUF];	    /* str; used by strcspn() */
char inputMemberSep[SHORT_BUF];	    /* str; used by strtok() */
char linkQuoteL[SHORT_BUF];
char linkQuoteR[SHORT_BUF];
char noLink[SHORT_BUF];
char outputFilenameSep;
char outputMemberSep;
UMAXTYPE__T PMASK =
    S_IRWXU | S_IRWXG | S_IRWXO | S_ISUID | S_ISGID | S_ISVTX;
char *Prog;
static char c_RCSid[] =
    "$Id: lstat.c,v 1.71 2012/01/05 03:05:28 adamm Exp adamm $";

#if defined(OS_SOLARIS)
    extern char *optarg;
    extern int optind, opterr, optopt;
#endif


int main(int argc, char *argv[])
{
Member M[SHORT_BUF];
int mm;
    int c;
    int i;
    int gotFileQuote = 0;
    int gotLinkQuote = 0;
    int gotMembers = 0;
    char members[LONG_BUF];
    char membersToo[LONG_BUF] = "";
    char mlist[MAX_MEMBERS][SHORT_BUF];
    int numMembers = 0;
    char *p;
    struct stat statbuf;
    int wantFileQuote = 0;
    int wantLinkQuote = 0;
    int xval = 0;

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	Prog = &argv[0][0];
    } else {
	++Prog;
    }

    /* Deal with the command line */
    if ((argc == 2) && (strcmp(argv[1], "--man") == 0)) {
	manpage();
	exit(0);
    }

    /* Fill in the defaults */
    dfltStrs();
    dfltMembers(members);
    setSeparators(
	"",
	&outputMemberSep, &outputFilenameSep,
	inputMemberSep, inputFormatSep
    );

    while ((c = getopt(argc, argv, "F:fhL:lMmN:O:o:s:v")) != -1) {
	switch ((char) c) {
	case 'F':
	    splitQuotes(optarg, fileQuoteL, fileQuoteR);
	    gotFileQuote = 1;
	    break;

	case 'f':
	    wantFileQuote = 1;
	    break;

	case 'h':	/* Yes, this overrides -o; so be it */
	    dfltMembersHuman(members);
	    break;

	case 'L':
	    splitQuotes(optarg, linkQuoteL, linkQuoteR);
	    gotLinkQuote = 1;
	    break;

	case 'l':
	    wantLinkQuote = 1;
	    break;

	case 'M':
	    showHumanMembers();
	    exit(0);

	case 'm':
	    showDfltMembers();
	    exit(0);

	case 'N':
	    (void) strcpy(noLink, optarg);
	    break;

	case 'O':
	    if (gotMembers) {
		fprintf(stderr,
		    "%s: -O and -o are mutually exclusive\n\n", Prog);
		usage();
		exit(1);
	    } else {
		++gotMembers;
	    }

	    (void) strcpy(membersToo, optarg);
	    break;

	case 'o':
	    if (gotMembers) {
		fprintf(stderr,
		    "%s: -O and -o are mutually exclusive\n\n", Prog);
		usage();
		exit(1);
	    } else {
		++gotMembers;
	    }

	    (void) strcpy(members, optarg);
	    break;

	case 's':
	    setSeparators(
		optarg,
		&outputMemberSep, &outputFilenameSep,
		inputMemberSep, inputFormatSep
	    );
	    break;

	case 'v':
	    (void) printf("%s\n", c_RCSid);
	    (void) printf("%s\n", h_RCSid);
	    exit(0);

        default:
            usage();
	    exit(1);
	}
    }

    /**
    *** If we get -L we shouldn't also get -l (and m.m. -F and -f), but
    *** just in case we get both, do the copying here and then only if
    *** we didn't get -L
    **/
    if ((wantLinkQuote == 1) && (gotLinkQuote == 0)) {
	(void) strcpy(linkQuoteL, "\"");
	(void) strcpy(linkQuoteR, "\"");
    }

    if ((wantFileQuote == 1) && (gotFileQuote == 0)) {
	(void) strcpy(fileQuoteL, "\"");
	(void) strcpy(fileQuoteR, "\"");
    }

    /* Is there anything left on the command line? */
    if (argc == optind) {
        usage();
	exit(1);
    }

    /**
    *** Because -h can replace the default list of members, wait
    *** until now to tack on anything we got from -O
    **/
    if (strlen(membersToo) > 0) {
	(void) strcat(members, inputMemberSep);
	(void) strcat(members, membersToo);
    }

    /**
    *** Split the string of members/modifiers into individual Member
    *** structs, check for errors, and fill in stuff like base, case,
    *** time format, etc.
    **/
    if (processMembers(members,M,&mm) > 0) {
	(void) fprintf(stderr, "%s: aborting.\n", Prog);
	exit(1);
    }

    /* Work our way through the list of files */
    for( ; optind < argc; optind++) {
	c = lstat(argv[optind], &statbuf);

	if (c == -1) {
	    (void) fprintf(stderr,
		"%s:%d: Unable to lstat() file '%s': %s\n",
		Prog, errno, argv[optind], strerror(errno));
	    xval = 2;
	    continue;
	}

	xval = showMembers(argv[optind], &statbuf, M, mm);
	(void) printf("%c%s%s%s\n",
		    outputFilenameSep,
		    fileQuoteL, argv[optind], fileQuoteR);
    }

    exit(xval);
} /* main */


void dfltMembers(char *members)
{
    (void) strcpy(members,
	"dev"
    	";ino"
    	";mode"
    	";nlink"
    	";uid"
    	";gid"
    	";size"
    	";atime"
    	";mtime"
    	";ctime"
    );
} /* dfltMembers */


void dfltMembersHuman(char *members)
{
    (void) strcpy(members,
    	"ino"
	";type:s"
	";perm:s"
    	";nlink"
    	";uid:s"
    	";gid:s"
    	";size"
    	";atime:z"
    	";mtime:z"
    	";ctime:z"
    );
} /* dfltMembersHuman */


void dfltStrs(void)
{
    fileQuoteL[0] = '\0';
    fileQuoteR[0] = '\0';
    linkQuoteL[0] = '\0';
    linkQuoteR[0] = '\0';

    strcpy(noLink, "//");
} /* dfltStrs */


int formatPermsChmod(
    char *buf,
    size_t buflen,
    mode_t mode,
    int maskR,
    int maskW,
    int maskX,
    int maskS,
    char *who,
    char *symbols,
    int comma
) {
    int b;
    int bits[4];
    char temp[SHORT_BUF];

    if (buflen < (size_t) MAX_PERM_SUBSTR) {
	return 3;
    }

    if (comma) {
	(void) snprintf(buf, buflen, ",%s", who);
    } else {
	(void) snprintf(buf, buflen, "%s", who);
    }

    /* Special handling for all bits == 0 */
    if ((mode & (maskR | maskW | maskX | maskS)) == 0) {
	(void) snprintf(temp, SHORT_BUF, "-%s", symbols);
	(void) strncat(buf, temp, buflen - strlen(buf));
	return 0;
    } else {
	(void) strncat(buf, "=", buflen - strlen(buf));
    }

    bits[0] = maskR;
    bits[1] = maskW;
    bits[2] = maskX;
    bits[3] = maskS;

    for (b = 0; b < 4; b++) {
	if (mode & bits[b]) {
	    (void) snprintf(temp, SHORT_BUF, "%c", symbols[b]);
	    (void) strncat(buf, temp, buflen - strlen(buf));
	}
    }

    return 0;
} /* formatPermsChmod */


int formatPermsLs(
    char *buf,
    size_t buflen,
    mode_t mode,
    int maskR,
    int maskW,
    int maskX,
    int maskS,
    char *symbols
) {

    if (buflen < (size_t) MAX_LS_SUBSTR) {
	return 3;
    }

    /* Special handling for all bits == 0 */
    if ((mode & (maskR | maskW | maskX | maskS)) == 0) {
	(void) strncpy(buf, "---", buflen);
	return 0;
    }

    /* Read */
    if (mode & maskR) {
	(void) strcpy(buf, "r");
    } else {
	(void) strcpy(buf, "-");
    }

    /* Write */
    if (mode & maskW) {
	(void) strcat(buf, "w");
    } else {
	(void) strcat(buf, "-");
    }

    /* Execute and "those other bits" */
    if (mode & maskX) {
	if (mode & maskS) {
	    (void) strncat(buf, symbols, 1);
	} else {
	    (void) strcat(buf, "x");
	}
    } else {
	if (mode & maskS) {
	    (void) strncat(buf, symbols+1, 1);
	} else {
	    (void) strcat(buf, "-");
	}
    }

    return 0;
} /* formatPermsLs */


void manpage(void)
{
    unsigned char *lines[] = {
	#include "lstat.man.hex"
	0
    };

    unhexify(lines);
} /* manpage */


int processMembers(char *string, Member *members, int *numMembers)
{
    int base;
    char buf[LONG_BUF];
    char *bp;
    size_t memlen;
    int modtype;
    int mx;
    int ret;
    int temp;
    size_t toklen;

    /**
    *** Because strtok modifies its args, make a copy of the string
    *** that contains the list of members and modifiers.
    **/
    if (strlen(string) < LONG_BUF) {
	strncpy(buf, string, LONG_BUF);
    } else {
	(void) fprintf(stderr,
	    "%s: member list too long: '%s'\n", Prog, string);
	exit(1);
    }

    mx = ret = 0;
    bp = strtok(buf, inputMemberSep);
    while (bp != NULL) {
	/* Check the number of members */
	if (mx > MAX_MEMBERS) {
	    (void) fprintf(stderr, "%s: too many members; aborting.\n",
		Prog);
	    exit(1);
	}

	/* Check the length of the member name */
	memlen = strcspn(bp, inputFormatSep);
	if (memlen > SHORT_BUF) {
	    (void) fprintf(stderr,
		"%s: member name too long: '%s'\n", Prog, bp);
	    (void) fprintf(stderr, "%s: aborting.\n", Prog);
	    exit(1);
	}

	/* Check the length of the modifier (if we got one) */
	toklen = strlen(bp);
	if (memlen < toklen) {
	    if ((toklen - memlen - 1) > SHORT_BUF) {
		(void) fprintf(stderr,
		    "%s: modifier too long: '%s'\n", Prog, bp);
		(void) fprintf(stderr, "%s: aborting.\n", Prog);
		exit(1);
	    }
	}

	/* Fill in the struct */
	(void) strnsubstr((char *) members[mx].name, bp, 0, memlen);
	if (memlen < toklen) {
	    (void) strnsubstr((char *) members[mx].modifier,
		bp, memlen + 1, toklen - memlen - 1);
	} else {
	    (void) strncpy((char *) members[mx].modifier, "", SHORT_BUF);
	}

	/* Check member name, valid modifiers */
	temp = validateMember(
	    (char *) members[mx].name, (char *) members[mx].modifier,
	    &members[mx].modtype, &members[mx].base);
	if (temp > 0) {
	    ret = 1;
	} else {
	    setBaseCase(
		(char *) members[mx].modifier, members[mx].modtype,
		&members[mx].base, &members[mx].uc
	    );
	    setFormat(
		(char *) members[mx].modifier,
		members[mx].modtype,
		(char *) members[mx].format
	    );
	}

	/* Next member in the list */
	++mx;
	bp = strtok(NULL, inputMemberSep);
    }

    *numMembers = mx;
    return ret;
} /* processMembers */


void setBaseCase(
    char *modifier,
    int modtype,
    int *base,
    int *uc
) {
    if (*base > 0) {
	*uc = 0;
	return;
    }

    if (modtype != MOD_NUMERIC) {
	return;
    }

    switch (modifier[0]) {
    case 'b':	*base =  2; *uc = 0; break;
    case 'o':	*base =  8; *uc = 0; break;
    case 'x':	*base = 16; *uc = 0; break;
    case 'X':	*base = 16; *uc = 1; break;
    default:
	(void) fprintf(stderr,
	    "%s: invalid modifier: '%s'; how did this happen?\n",
	    Prog, modifier);
	exit(3);
	break;
    }

    return;
} /* setBaseCase */


void setFormat(
    char *modifier,
    int modtype,
    char *format
) {
    if (modtype != MOD_DATETIME) {
	return;
    }

    switch (modifier[0]) {
    case 'd':	(void) strcpy(format, "%Y-%m-%d");		break;
    case 'D':	(void) strcpy(format, "%Y%m%d");		break;
    case 't':	(void) strcpy(format, "%Y-%m-%d-%H-%M-%S");	break;
    case 'T':	(void) strcpy(format, "%Y%m%d%H%M%S");		break;
    case 'z':	(void) strcpy(format, "%Y-%m-%d.%H:%M:%S");	break;
    case 'Z':	(void) strcpy(format, "%Y%m%d.%H%M%S");		break;
    default:
	(void) fprintf(stderr,
	    "%s: invalid modifier: '%s'; how did this happen?\n",
	    Prog, modifier);
	exit(3);
	break;
    }

    return;
} /* setFormat */


void setSeparators(
    char *str,			/* char[] */
    char *outputMemberSep,	/* char   */
    char *outputFilenameSep,	/* char   */
    char *inputMemberSep,	/* char[] */
    char *inputFormatSep	/* char[] */
) {
    int len;

    len = strlen(str);

    if (len > 0) {
	*outputMemberSep = *str;
    } else {
	*outputMemberSep = DFLT_OMSEP;
    }

    if (len > 1) {
	*outputFilenameSep = *(str + 1);
    } else {
	*outputFilenameSep = DFLT_OFNSEP;
    }

    /* Used by strtok(), so it has to be a string */
    if (len > 2) {
	(void) sprintf(inputMemberSep, "%c", *(str + 2));
    } else {
	(void) sprintf(inputMemberSep, "%c", DFLT_IMSEP);
    }

    /* Used by strcspn(), so it has to be a string */
    if (len > 3) {
	(void) sprintf(inputFormatSep, "%c", *(str + 3));
    } else {
	(void) sprintf(inputFormatSep, "%c", DFLT_IFSEP);
    }

    return;
} /* setSeparators */


void showDfltMembers(void)
{
    char members[LONG_BUF];

    dfltMembers(members);
    printf("%s\n", members);
} /* showDfltMembers */


void showHumanMembers(void)
{
    char members[LONG_BUF];

    dfltMembersHuman(members);
    printf("%s\n", members);
} /* showHumanMembers */


int showLsMode(char *buf, size_t buflen, mode_t mode)
{
    int ret;
    char temp[SHORT_BUF];

    if (buflen < (size_t) MAX_PERM_STR) {
	return 3;
    }

    /* File type */
    ret = showSymbolicType(temp, sizeof(temp), mode, '-');
    if (ret == 0) {
	(void) strncpy(buf, temp, buflen);
    } else {
	return ret;
    }

    /* User */
    ret = formatPermsLs(temp, sizeof(temp), mode,
	S_IRUSR, S_IWUSR, S_IXUSR, S_ISUID, "sS");
    if (ret == 0) {
	(void) strncat(buf, temp, buflen);
    } else {
	return ret;
    }

    /* Group */
    ret = formatPermsLs(temp, sizeof(temp), mode,
	S_IRGRP, S_IWGRP, S_IXGRP, S_ISGID, "sS");
    if (ret == 0) {
	(void) strncat(buf, temp, buflen - strlen(buf));
    } else {
	return ret;
    }

    /* Other */
    ret = formatPermsLs(temp, sizeof(temp), mode,
	S_IROTH, S_IWOTH, S_IXOTH, S_ISVTX, "tT");
    if (ret == 0) {
	(void) strncat(buf, temp, buflen - strlen(buf));
    } else {
	return ret;
    }

    return 0;
} /* showLsMode */


int showMembers(
    char *filename,
    struct stat *sbuf,
    Member *members,
    int numMembers
) {
    int base = 10;
    char buf[LONG_BUF];
    int m;
    int n;
    int rval;
    mode_t temp;

    for (m = 0; m < numMembers; m++) {
	rval = 0;

/* DO NOT #ifdef! */
	if (strcmp(members[m].name, "atime") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_atime,
				"time_t", members[m]);

#ifdef HAS__stBLKSIZE
	} else if (strcmp(members[m].name, "blksize") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_blksize,
				BLKSIZE__S, members[m]);
#endif

#ifdef HAS__stBLOCKS
	} else if (strcmp(members[m].name, "blocks") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_blocks,
				BLOCKS__S, members[m]);
#endif

/* DO NOT #ifdef! */
	} else if (strcmp(members[m].name, "ctime") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_ctime,
				"time_t", members[m]);

#ifdef HAS__stDEV
	} else if (strcmp(members[m].name, "dev") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_dev,
				DEV__S, members[m]);
#endif

#ifdef HAS__stFLAGS
	} else if (strcmp(members[m].name, "flags") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_flags,
				FFLAGS__S, members[m]);
#endif

#ifdef HAS__stGEN
	} else if (strcmp(members[m].name, "gen") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_gen,
				GEN__S, members[m]);
#endif

#ifdef HAS__stGID
	} else if (strcmp(members[m].name, "gid") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_gid,
				GID__S, members[m]);
#endif

#ifdef HAS__stINO
	} else if (strcmp(members[m].name, "ino") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_ino,
				INO__S, members[m]);
#endif

/* No need for #ifdef on this one */
	} else if (strcmp(members[m].name, "link") == 0) {
	    n = showTarget(filename, buf, sizeof(buf), sbuf->st_mode);

#ifdef HAS__stMODE
	} else if (strcmp(members[m].name, "mode") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_mode,
				MODE__S, members[m]);
#endif

/* DO NOT #ifdef! */
	} else if (strcmp(members[m].name, "mtime") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_mtime,
				"time_t", members[m]);

#ifdef HAS__stNLINK
	} else if (strcmp(members[m].name, "nlink") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_nlink,
				NLINK__S, members[m]);
#endif

#ifdef HAS__stMODE
	} else if (strcmp(members[m].name, "perm") == 0) {
	    temp = sbuf->st_mode & PMASK;
	    n = typeToStr(buf, sizeof(buf), (void *) &temp,
				MODE__S, members[m]);
#endif

#ifdef HAS__stRDEV
	} else if (strcmp(members[m].name, "rdev") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_rdev,
				RDEV__S, members[m]);
#endif

#ifdef HAS__stSIZE
	} else if (strcmp(members[m].name, "size") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_size,
				SIZE__S, members[m]);
#endif

#ifdef HAS__stMODE
	} else if (strcmp(members[m].name, "type") == 0) {
	    temp = sbuf->st_mode & ~PMASK;
	    n = typeToStr(buf, sizeof(buf), (void *) &temp,
				MODE__S, members[m]);
#endif

#ifdef HAS__stUID
	} else if (strcmp(members[m].name, "uid") == 0) {
	    n = typeToStr(buf, sizeof(buf), (void *) &sbuf->st_uid,
				UID__S, members[m]);
#endif

	} else {
	    (void) strcpy(buf, "?");
	    n = 1;
	}

	/* Show any errorrs from typeToStr() */
	if (n != 0) {
	    (void) sprintf(buf, "{%d}", n);
	}

	if (m > 0) {
	    (void) printf("%c", outputMemberSep);
	}

	(void) printf("%s", buf);

	if (n != 0) {
	    rval = n;
	}
    }

    return rval;
} /* showMembers */


int showNumeric(char *buf, size_t buflen, UMAXTYPE__T num, Member member)
{
    char digits[] = {
	'0', '1', '2', '3', '4', '5', '6', '7',
	'8', '9', 'a', 'b', 'c', 'd', 'e', 'f'
    };

    int i;
    int j;
    UMAXTYPE__T t;

    /* Make sure there's a null at the end of the string */
    i = buflen - 1;
    *(buf + i) = '\0';
    --i;

    /**
    *** Get the low-order digit and stuff it into the buffer;
    *** rinse, repeat until bored
    **/
    do {
	if (i < 0) {
	    return 3;
	}

	t = num % member.base;
	num = num / member.base;

	if ((member.uc != 0) && (t > 9)) {
	    *(buf + i) = toupper((int) digits[t]);
	} else {
	    *(buf + i) = digits[t];
	}

	--i;
    } while (num > 0);

    /* Shift the digits to the front of the string */
    if (i >= 0) {
	for (++i, j = 0; i < buflen; i++, j++) {
	    *(buf + j) = *(buf + i);
	}
    }

    return 0;
} /* showNumeric */


int showSymbolicGid(char *buf, size_t buflen, void *gid)
{
    GID__T *gidPtr;
    struct group *groupPtr;
    char temp[LONG_BUF];

    gidPtr = (GID__T *) gid;

    /**
    *** We have to get the group name before we can check to see
    *** if there's enough room in the buffer for the name.
    ***
    *** It's not really an error if we can't get a name for the
    *** group; if that happens, just show the group numerically.
    **/

    /* Get the group entry and check for error */
    groupPtr = getgrgid(*gidPtr);
    if (groupPtr != NULL) {
	strncpy(temp, groupPtr->gr_name, LONG_BUF);
    } else {
	(void) snprintf(temp, LONG_BUF, "%lu", *((GID__T *) gid));
    }

    /* Is there enough space? */
    if (buflen < strlen(temp)) {
	return 3;
    }

    /* Append the results to the buffer */
    strncpy(buf, temp, LONG_BUF);
    return 0;
} /* showSymbolicGid */


int showSymbolicPerm(char *buf, size_t buflen, mode_t perm)
{
    int ret;
    char temp[SHORT_BUF];

    if (buflen < (size_t) MAX_PERM_STR) {
	return 3;
    }

    /* User */
    ret = formatPermsChmod(temp, sizeof(temp), perm,
	S_IRUSR, S_IWUSR, S_IXUSR, S_ISUID, "u", "rwxs", 0);
    if (ret == 0) {
	(void) strncpy(buf, temp, buflen);
    } else {
	return ret;
    }

    /* Group */
    ret = formatPermsChmod(temp, sizeof(temp), perm,
	S_IRGRP, S_IWGRP, S_IXGRP, S_ISGID, "g", "rwxs", 1);
    if (ret == 0) {
	(void) strncat(buf, temp, buflen - strlen(buf));
    } else {
	return ret;
    }

    /* Other */
    ret = formatPermsChmod(temp, sizeof(temp), perm,
	S_IROTH, S_IWOTH, S_IXOTH, S_ISVTX, "o", "rwxt", 1);
    if (ret == 0) {
	(void) strncat(buf, temp, buflen - strlen(buf));
    } else {
	return ret;
    }

    return 0;
} /* showSymbolicPerm */


int showSymbolicType(char *buf, size_t buflen, mode_t mode, char file)
{
    if (buflen < (size_t) 2) {
	return 3;
    }

    switch (mode & S_IFMT) {
    case S_IFBLK:   (void) strncpy(buf, "b", buflen);   break;
    case S_IFCHR:   (void) strncpy(buf, "c", buflen);   break;
    case S_IFDIR:   (void) strncpy(buf, "d", buflen);   break;
    case S_IFIFO:   (void) strncpy(buf, "p", buflen);   break;
    case S_IFLNK:   (void) strncpy(buf, "l", buflen);   break;
    case S_IFREG:
	(void) snprintf(buf, buflen, "%c", file);
	break;
    case S_IFSOCK:  (void) strncpy(buf, "s", buflen);   break;
    default:	(void) strncpy(buf, "?", buflen);   break;
    }

    return 0;
} /* showSymbolicType */


int showSymbolicTypePerm(char *buf, size_t buflen, mode_t mode)
{
    if ((mode & S_IFMT) > 0) {
	return showSymbolicType(buf, buflen, mode, 'f');
    } else {
	return showSymbolicPerm(buf, buflen, mode);
    }
} /* showSymbolicTypePerm */


int showSymbolicUid(char *buf, size_t buflen, void *uid)
{
    UID__T *uidPtr;
    struct passwd *passwdPtr;
    char temp[LONG_BUF];

    uidPtr = (UID__T *) uid;

    /**
    *** We have to get the user name before we can check to see
    *** if there's enough room in the buffer for the name.
    ***
    *** It's not really an error if we can't get a name for the
    *** group; if that happens, just show the group numerically.
    **/

    /* Get the passwd entry and check for error */
    passwdPtr = getpwuid(*uidPtr);
    if (passwdPtr != NULL) {
	strncpy(temp, passwdPtr->pw_name, LONG_BUF);
    } else {
	(void) snprintf(temp, LONG_BUF, "%lu", *((UID__T *) uid));
    }

    /* Is there enough space? */
    if (buflen < strlen(temp)) {
	return 3;
    }

    /* Append the results to the buffer */
    (void) strncpy(buf, temp, LONG_BUF);
    return 0;
} /* showSymbolicUid */


int showTarget(char *file, char *buf, size_t buflen, mode_t mode)
{
    int b;
    char path[LONG_BUF];
    size_t s;

    /* Quick size check */
    if (buflen < (size_t) 3) {
	return 3;
    }

    /* If the file isn't a symlink, show noLink[] */
    if (S_ISLNK(mode) == 0) {
	(void) sprintf(buf, "%s%s%s", linkQuoteL, noLink, linkQuoteR);
	return 0;	/* this is NOT an error! */
    }

    /**
    *** We have to get the target and find its length before we can see
    *** if the caller gave us a big enough buffer; we also need to deal
    *** with the case of readlink() failing -- which we'll do first to
    *** keep the code even cleaner.
    **/
    (void) memset(path, 0, LONG_BUF);
    if ((int) readlink(file, path, LONG_BUF) == -1) {
	return 3;
    }

    /* Check the size of the buffer (again) */
    s = strlen(path) + strlen(linkQuoteL) + strlen(linkQuoteR) + 1;
    if (buflen < s) {
	return 3;
    }

    /* Get on with it */
    (void) sprintf(buf, "%s%s%s", linkQuoteL, path, linkQuoteR);
    return 0;
} /* showTarget */


void splitQuotes(char *buf, char *left, char *right)
{
    int h;

    (void) memset(left, 0, SHORT_BUF);
    (void) memset(right, 0, SHORT_BUF);

    h = strlen(buf) / 2;

    (void) strncpy(left, buf, h);
    (void) strncpy(right, buf + h, h);
} /* splitQuotes */


char *strnsubstr(
    char *s1,
    const char *s2,
    size_t start,
    size_t n
) {
    (void) strncpy(s1, (s2 + start), n);
    *(s1 + n) = '\0';
    return s1;
} /* strnsubstr */


int typeToStr(
    char *buf,
    size_t buflen,
    void *valptr,
    char *valtype,
    Member member
) {
    UMAXTYPE__T n;
    struct tm *tbuf;

    /* Jump through flaming hoops to get the value */
    if (strcmp(valtype, "long") == 0) {
	n = (UMAXTYPE__T) (*(long *)valptr);

    } else if (strcmp(valtype, "unsigned long") == 0) {
	n = (UMAXTYPE__T) (*(unsigned long *)valptr);

#ifdef HAS__BLKCNT_T
    } else if (strcmp(valtype, BLKCNT__S) == 0) {
	n = (UMAXTYPE__T) (*(BLKCNT__T *)valptr);
#endif

#ifdef HAS__BLKSIZE_T
    } else if (strcmp(valtype, BLKSIZE__S) == 0) {
	n = (UMAXTYPE__T) (*(BLKSIZE__T *)valptr);
#endif

#ifdef HAS__DEV_T
    } else if (strcmp(valtype, DEV__S) == 0) {
	n = (UMAXTYPE__T) (*(DEV__T *)valptr);
#endif

#ifdef HAS__FFLAGS_T
    } else if (strcmp(valtype, FFLAGS__S) == 0) {
	n = (UMAXTYPE__T) (*(FFLAGS__T *)valptr);
#endif

#ifdef HAS__GID_T
    } else if (strcmp(valtype, GID__S) == 0) {
	n = (UMAXTYPE__T) (*(GID__T *)valptr);
#endif

#ifdef HAS__INO_T
    } else if (strcmp(valtype, INO__S) == 0) {
	n = (UMAXTYPE__T) (*(INO__T *)valptr);
#endif

#ifdef HAS__INT64_T
    } else if (strcmp(valtype, "int64_t") == 0) {
	n = (UMAXTYPE__T) (*(int64_t *)valptr);
#endif

#ifdef HAS__MODE_T
    } else if (strcmp(valtype, MODE__S) == 0) {
	n = (UMAXTYPE__T) (*(MODE__T *)valptr);
#endif

#ifdef HAS__NLINK_T
    } else if (strcmp(valtype, NLINK__S) == 0) {
	n = (UMAXTYPE__T) (*(NLINK__T *)valptr);
#endif

#ifdef HAS__OFF_T
    } else if (strcmp(valtype, OFF__S) == 0) {
	n = (UMAXTYPE__T) (*(OFF__T *)valptr);
#endif

#ifdef HAS__QUAD_T
    } else if (strcmp(valtype, "quad_t") == 0) {
	n = (UMAXTYPE__T) (*(quad_t *)valptr);
#endif

#ifdef HAS__TIME_T
    } else if (strcmp(valtype, XTIME__S) == 0) {
	n = (UMAXTYPE__T) (*(XTIME__T *)valptr);
#endif

#ifdef HAS__U_INT32_T
    } else if (strcmp(valtype, "u_int32_t") == 0) {
	n = (UMAXTYPE__T) (*(u_int32_t *)valptr);
#endif

#ifdef HAS__UINT32_T
    } else if (strcmp(valtype, "uint32_t") == 0) {
	n = (UMAXTYPE__T) (*(uint32_t *)valptr);
#endif

#ifdef HAS__U_LONG
    } else if (strcmp(valtype, "u_long") == 0) {
	n = (UMAXTYPE__T) (*(u_long *)valptr);
#endif

#ifdef HAS__UID_T
    } else if (strcmp(valtype, UID__S) == 0) {
	n = (UMAXTYPE__T) (*(UID__T *)valptr);
#endif

    } else {
	return 3;
    }

    /* Decide how to display the member */
    switch (member.modtype) {
    case MOD_NUMERIC:
	return showNumeric(buf, buflen, n, member);

    case MOD_DATETIME:
	tbuf = localtime((time_t *) &n);
	if (strftime(buf, buflen, member.format, tbuf) == 0) {
	    return 3;
	} else {
	    return 0;
	}

    case MOD_LS:
	return showLsMode(buf, buflen, (mode_t) n);

    case MOD_SYMBOLIC:
	if (strcmp(valtype, GID__S) == 0) {
	    return showSymbolicGid(buf, buflen, &n);
	} else if (strcmp(valtype, MODE__S) == 0) {
	    return showSymbolicTypePerm(buf, buflen, (mode_t) n);
	} else if (strcmp(valtype, UID__S) == 0) {
	    return showSymbolicUid(buf, buflen, &n);
	}

    default:
	/* How did this happen? */
	(void) fprintf(stderr,
	    "%s: invalid modtype: '%d'; how did this happen? Aborting.\n",
	    Prog, member.modtype);
	exit(3);
	break;
    }
} /* typeToStr */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }
} /* unhexify */


void usage()
{
    (void) fprintf(stderr, "Usage: %s %s\n",
		Prog, "[ options ] file [ ... ]");
    (void) fprintf(stderr, "\nUse \"%s --man\" for more details.\n",
		Prog);
} /* usage */


int validateMember(char *name, char *modifier, int *modtype, int *base)
{
    int m;
    char modDateTime[] = "dDtTzZ";
    char modNumeric[] = "boxX";
    int numValid;
    char *sp;

    /* Look it up in the table of known members */
    numValid = (int) sizeof(ValidMembers) / sizeof(Member);
    for (m = 0; m < numValid; m++) {
	if (strcmp(name, ValidMembers[m].name) == 0) {
	    break;
	}
    }

    if (m == numValid) {
	(void) fprintf(stderr, "%s: unknown member: '%s'\n", Prog, name);
	return 1;
    }

    /* Determine the type of modifier */
    if (modifier[0]) {
	if (strchr(modNumeric, modifier[0]) != NULL) {
	    *modtype = MOD_NUMERIC;
	} else if (strchr(modDateTime, modifier[0]) != NULL) {
	    *modtype = MOD_DATETIME;
	    *base = 0;
	} else if (modifier[0] == 'L') {
	    *modtype = MOD_LS;
	    *base = 0;
	} else if (modifier[0] == 's') {
	    *modtype = MOD_SYMBOLIC;
	    *base = 0;
	} if (isdigit((int) modifier[0])) {
	    /* If it's an actual numeric base, check the range */
	    *modtype = MOD_NUMERIC;
	    *base = atoi(modifier);
	    if ((*base < 2) || (*base > 16)) {
		(void) fprintf(stderr,
		    "%s: base '%d' out of range (2 - 16): '%s%c%s'\n",
		    Prog, *base, name, *inputFormatSep, modifier);
		return 1;
	    }
	}

	/* Test against the allowed types and return on error */
	if ((ValidMembers[m].modtype & *modtype) == 0) {
	    (void) fprintf(stderr,
		"%s: invalid modifier '%s': '%s%c%s'\n",
		Prog, modifier, name, *inputFormatSep, modifier);
	    return 1;
	}

    } else {
	*modtype = MOD_NUMERIC;
	*base = 10;
    }

    return 0;
} /* validateMember */


/*	===== MAN PAGE STARTS HERE =====
NAME
    lstat -- show the results of lstat(2)

SYNOPSIS
    lstat [ options ] file [ ... ]

BRIEF DESCRIPTION
    lstat calls lstat(2) on each file specified and prints the contents
    of the stat struct. By default, all values are shown in base 10;
    this can be changed, to either a different base or a symbolic
    representation (where appropriate), by using a modifier as described
    below.

OPTIONS
    -F str
	quote the filename using "str"; implies -f

    -f
	quote the filename

    -h
	show the default list of members (plus type and perm) with mode,
	type, perm, uid, gid, atime, mtime, and ctime shown in either an
	alternate base or symbolically; overrides -o but may be used
	with -O

    -L str
	quote the symbolic link target using "str"; implies -l

    -l
	quote the symbolic link target

    -m
	show the default list of member names then exit

    -M
	show the list of members produced by "-h" then exit

    -N str
	if "link" is specified in the list of members but the file is
	not a symbolic link, show "str" for the link target (default
	is "//")

    -O str
	a comma-separated list of members to be appended to the default
	list of members; can be used with -o or -h

    -o str
	a comma-separated list of members to be displayed

    -s str
	see "Separators" (below)

    -v
	show the version number then exit

FULL DESCRIPTION
    lstat allows the user to specify the files to be reported on, the
    members of the stat struct to be shown, the format of the output,
    and the format of the input list of members.

    Member names are the same as in the stat struct but without the
    leading "st_". Not all members of the stat struct are supported on
    every architecture; the following members should always be
    available: dev, ino, mode, nlink, uid, gid, size, atime, mtime, and
    ctime. The default output shows the preceeding list of members, in
    the order specified, with values separated by semicolons (";"); the
    name of the file is then shown, separated by a space. The -O flag
    appends user-specified members to the preceeding list.

    Several "synthetic" member names are also available:

	perm
	    the file permissions (the low-order portion of "mode")

	type
	    the file type (the high-order portion of "mode")

	link
	    if the file is a symbolic link, the target; if the file is
	    not a symlink, the string "//"; the -? flag can be used to
	    specify an alternate string

    By default, all members are shown in base 10; this can be changed by
    specifying the member name, followed by a colon and a base (2 - 16)
    or one of the following letters:

	b    binary
	o    octal
	x    hexadecimal (lower case letters)
	X    hexadecimal (upper case letters)

    For atime, ctime, and mtime, the following letters are also
    recognized:

	d    YYYY-MM-DD
	D    YYYYMMDD
	t    YYYY-MM-DD-hh-mm-ss
	T    YYYYMMDDhhmmss
	z    YYYY-MM-DD.hh:mm:ss
	Z    YYYYMMDD.hhmmss

    For gid, perm, type, and uid the modifier "s" may be used to show
    the member in "symbolic" form. For gid and uid, this will be the
    name of the group or user; for perm, the output will be a comma-
    separated string in the form understood by chmod(1); for type, the
    output will be the same letter as the first character of the output
    from "ls -l" except that "f" will be used for plain files (ls would
    show "-").

    For mode, the modifier "L" may also be used; this will show the same
    output as ls(1). (It would have been more mnemonic to have used
    lower case ell, but that's too easily confused with the digit one.)

    If the -h flag is given, the default member list will be as follows
    (split for readability):

	"dev,ino,mode:o,type:s,perm:s,nlink,"
	"uid,gid,size,atime:z,mtime:z,ctime:z"

    In the output, the file name and symbolic link target can be shown
    surrounded by quotes. By default, the double-quote character is
    used; this can be changed with the -F switch (for the file name) or
    the -L switch (for the target). Both switches accept a string as an
    argument; the first half of the string will be displayed before the
    member, the second half after the member. If the string is an odd
    number of characters, the behavior is unspecified; each string may
    be up to 64 characters long.

    Separators
	lstat uses four separators (2 for input, 2 for output) as
	follows:

	    output member separator
		in the output, the separator between members;
		default is semicolon (";") because "perm:s"
		contains commas (",")

	    output filename separator
		in the output, the separator between the last
		member and the file name; default is space (" ")

	    input member separator
		in conjunction with -o, the separator between
		members; default is semicolon (";")

	    input format separator
		also in conjunction with -o, the separator
		between a member and the output format
		specifier (a numeric base, a letter indicating a
		base, or a letter indicating a 'symbolic" format);
		default is colon (":")

	Any or all of these separators can be changed with the -s flag.
	This flag takes one to four characters, corresponding to the
	list of separators above, in the order shown above. For example:

	    lstat -s "." -o uid,gid somefile

	gives:

	    101.101 somefile

	Another example:

	    lstat -s "|:.=" -o "uid=s.mtime=T" somefile

	gives:

	    adamm|20090911094249:somefile

	Changing the output separators is always safe. Changing the
	input separators should be used only in conjunction with the -o
	flag; changing the input separators with -O, -h, or the default
	member list will almost certainly break things.

EXIT STATUS
    lstat will exit with one of the following values:

	0    no errors
	1    you did something wrong
	2    lstat() failed
	3    something else failed

    There are lots of causes of each type of error; details will be
    printed on standard error.

BUGS
    Several. :-)

EXAMPLE
    lstat somefile
    2049;1734;33188;1;101;101;459;1252677159;1252676569;1252676569 somefile

    lstat -o dev,ino,mode,nlink,uid,gid,size,atime,mtime,ctime somefile
    2049;1734;33188;1;101;101;459;1252677159;1252676569;1252676569 somefile

    lstat -o "type:s,perm:o,perm:s,uid,uid:s" somefile
    f;644;u=rw,g=r,o=r;101;adamm somefile

    lstat -o 'mtime:z' somefile
    2009-09-11.09:42:49 somefile

SEE ALSO
    lstat(2), chmod(1), ls(1)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====		*/


/**
*** Copyright, 2009, 2010, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
