/**
***	$RCSfile: lstat.h,v $
***	$Revision: 1.19 $
***	$Date: 2010/08/14 02:42:57 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


static char	h_RCSid[] =
    "$Id: lstat.h,v 1.19 2010/08/14 02:42:57 adamm Exp adamm $";


/* Sizes */
#define LONG_BUF		8192
#define SHORT_BUF		64
#define MAX_LS_STR		11	/* "-rwxrwxrwx" + '\0' */
#define MAX_LS_SUBSTR		4	/* "rwx" + '\0' */
#define MAX_MEMBERS		64
#define MAX_PERM_STR		21	/* "u=rwxs,g=rwxs,o=rwxt" + '\0' */
#define MAX_PERM_SUBSTR		8	/* ",?=rwx?" + '\0' */

/* Default separators, always as char! */
#define DFLT_IFSEP		':'
#define DFLT_IMSEP		';'
#define DFLT_OFNSEP		' '
#define DFLT_OMSEP		';'

/* Used by strchr, so CHR_...! */
#define FFS			CHR_COLON

/* Used by strtok, so STR_... */
#define IFS			STR_COMMA
#define OFS			STR_COMMA

/* Largest unsigned type */
#define UMAXTYPE__T	uintmax_t
    /**
    *** Might need to be unsigned long or unsigned long long
    **/


/* Members */
#define MOD_NUMERIC	1
#define MOD_DATETIME	2
#define MOD_SYMBOLIC	4
#define MOD_LS		8

typedef struct {
    char name[SHORT_BUF];
    int modtype;
    int modifier[SHORT_BUF];
    int base;
    int uc;
    char format[SHORT_BUF];
} Member;


Member ValidMembers[] = {
    { "atime",	    MOD_NUMERIC | MOD_DATETIME },
    { "blksize",    MOD_NUMERIC                },
    { "blocks",	    MOD_NUMERIC                },
    { "ctime",	    MOD_NUMERIC | MOD_DATETIME },
    { "dev",	    MOD_NUMERIC                },
    { "flags",	    MOD_NUMERIC                },
    { "gen",	    MOD_NUMERIC                },
    { "gid",	    MOD_NUMERIC | MOD_SYMBOLIC },
    { "ino",	    MOD_NUMERIC                },
    { "link",	    0                          },
    { "mode",	    MOD_NUMERIC | MOD_LS       },
    { "mtime",	    MOD_NUMERIC | MOD_DATETIME },
    { "nlink",	    MOD_NUMERIC                },
    { "perm",	    MOD_NUMERIC | MOD_SYMBOLIC },
    { "rdev",	    MOD_NUMERIC                },
    { "size",	    MOD_NUMERIC                },
    { "type",	    MOD_NUMERIC | MOD_SYMBOLIC },
    { "uid",	    MOD_NUMERIC | MOD_SYMBOLIC },
};


/* Defaults */
#define HAS__BLKCNT_T
#define HAS__DEV_T
#define HAS__GID_T
#define HAS__INO_T
#define HAS__MODE_T
#define HAS__NLINK_T
#define HAS__OFF_T
#define HAS__TIME_T
#define HAS__UID_T


#define HAS__stBLOCKS
#define HAS__stDEV
#define HAS__stGID
#define HAS__stINO
#define HAS__stMODE
#define HAS__stNLINK
#define HAS__stSIZE
#define HAS__stUID
#define HAS__stXTIME

/**
*** Just because a default type is defined does NOT necessarily mean
*** that the corresponding member exists; these are here simply to
*** reduce the number of #defines we need further down
**/
#define BLKCNT__T	 blkcnt_t
#define BLKCNT__S	"blkcnt_t"
#define BLKSIZE__T	 blksize_t
#define BLKSIZE__S	"blksize_t"
#define BLOCKS__T	 blkcnt_t
#define BLOCKS__S	"blkcnt_t"
#define DEV__T		 dev_t
#define DEV__S		"dev_t"
#define FFLAGS__T	 fflags_t
#define FFLAGS__S	"fflags_t"
#define GEN__T		 uint32_t
#define GEN__S		"uint32_t"
#define GID__T		 gid_t
#define GID__S		"gid_t"
#define INO__T		 ino_t
#define INO__S		"ino_t"
#define MODE__T		 mode_t
#define MODE__S		"mode_t"
#define NLINK__T	 nlink_t
#define NLINK__S	"nlink_t"
#define OFF__T	 	 off_t
#define OFF__S		"off_t"
#define SIZE__T		 off_t
#define SIZE__S		"off_t"
#define XTIMENSEC__T	 long
#define XTIMENSEC__S	"long"
#define XTIMESPEC__T	 struct timespec
#define XTIMESPEC__S	"struct timespec"
#define UID__T		 uid_t
#define UID__S		"uid_t"
#define XTIME__T	 time_t
#define XTIME__S	"time_t"
#define UINT32__T	 uint32_t
#define UINT32__S	"uint32_t"
#define INT64__T	 int64_t
#define INT64__S	"int64_t"


/**
*** Old RedHat 7 stuff
***
***	#define BLKSIZE__T		 unsigned long
***	#define BLKSIZE__S		"unsigned long"
***	#define BLOCKS__T		 unsigned long
***	#define BLOCKS__S		"unsigned long"
**/

#ifdef OS_FREEBSD
#define HAS__UINT32_T

#define HAS__stBIRTHTIMEX
#define HAS__stRDEV
#define HAS__stGEN
#define HAS__stXTIMENSEC
#define HAS__stXTIMESPEC

#undef BLKSIZE__T
#undef BLKSIZE__S
#undef BLOCKS__T
#undef BLOCKS__S

#define BLKSIZE__T		UINT32__T
#define BLKSIZE__S		UINT32__S
#define BLOCKS__T		INT64__T
#define BLOCKS__S		INT64__S
#define RDEV__T		DEV__T
#define RDEV__S		DEV__S
#endif

#ifdef OS_IRIX
#undef HAS__stXTIME
#define HAS__stXTIM
#endif



/**
*** HAS__BLKCNT_T
*** HAS__BLKSIZE_T
*** HAS__DEV_T
*** HAS__FFLAGS_T
*** HAS__GID_T
*** HAS__INT64_T
*** HAS__INO64_T
*** HAS__INO_T
*** HAS__INT64_T
*** HAS__MODE_T
*** HAS__NLINK_T
*** HAS__OFF64_T
*** HAS__OFF_T
*** HAS__QUAD_T
*** HAS__TIME_T
*** HAS__TIMESPEC_T
*** HAS__U_INT32_T
*** HAS__UINT32_T
*** HAS__UINTMAX_T
*** HAS__U_LONG
*** HAS__UID_T
**/


/**
*** HAS__stXTIM
*** HAS__stXTIME
*** HAS__stXTIMENSEC
*** HAS__stXTIMESPEC
*** HAS__stBIRTHTIMEX
*** HAS__stBLKSIZE
*** HAS__stBLOCKS
*** HAS__stDEV
*** HAS__stFLAGS
*** HAS__stGEN
*** HAS__stGID
*** HAS__stINO
*** HAS__stMODE
*** HAS__stNLINK
*** HAS__stRDEV
*** HAS__stSIZE
*** HAS__stUID
**/



/**
*** Copyright, 2009-2010, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
