/**
***	$RCSfile: mkstemp.c,v $
***	$Revision: 1.4 $
***	$Date: 2009/12/11 13:01:33 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


static char RCSid[] =
    "$Id: mkstemp.c,v 1.4 2009/12/11 13:01:33 adamm Exp $";


void manpage(void);
void unhexify(unsigned char **);
void usage(char *);

char *Prog;


int main(int argc, char *argv[])
{
    int	    fd;
    char    filename[8192];
    char    message[BUFSIZ];

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	Prog = &argv[0][0];
    } else {
	++Prog;
    }

    /* Deal with --help and --man */
    if (argc > 1) {
	if (
	    (strcmp(argv[1], "--help") == 0)
	    ||
	    (strcmp(argv[1], "--man") == 0)
	) {
	    manpage();
	    exit(0);
	}
    }

    /* Exactly one arg -- the file name template */
    if (argc != 2) {
	(void) sprintf(message, "expected 1 arg; got %d.", argc - 1);
	usage(message);
	exit(1);
    }

    /* Make sure there are at least six "X"s */
    if (strstr(argv[1], "XXXXXX") == NULL) {
	(void) sprintf(message, "invalid template: '%s'", argv[1]);
	usage(message);
	exit(1);
    }

    /* Save it */
    (void) strcpy(filename, argv[1]);

    fd = mkstemp(filename);
    if (fd == -1) {
	(void) fprintf(stderr,
		    "%s: mkstemp(\"%s\"): errno = %d\n",
		    Prog, filename, errno);
	(void) perror(Prog);
	exit(1);
    }

    (void) close(fd);
    (void) printf("%s\n", filename);
    exit(0);
} /* main */


void manpage(void)
{
    unsigned char *lines[] = {
	#include "mkstemp.man.hex"
	0
    };

    unhexify(lines);
    return;
} /* manpage */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }

    return;
} /* unhexify */


void usage(char *message)
{
    (void) fprintf(stderr, "%s: %s\n\n", Prog, message);
    (void) fprintf(stderr, "Use \"%s --man\" for more details.\n", Prog);
    return;
} /* usage */


/*	===== MAN PAGE STARTS HERE =====
NAME
    mkstemp -- command-line interface to mkstemp(3)

SYNOPSIS
    mkstemp template

DESCRIPTION
    mkstemp passes "template" to mkstemp(3) then prints the name of the
    file that is created.

    "template" should look like a filename with six trailing 'X's.

EXIT STATUS
    mkstemp will exit with one of the following values:

	0    success
	1    error

SEE ALSO
    mkstemp(3)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====	*/


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
