/**
***	$RCSfile: numcmp.c,v $
***	$Revision: 1.3 $
***	$Date: 2011/08/07 13:02:45 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


static char RCSid[] =
    "$Id: numcmp.c,v 1.3 2011/08/07 13:02:45 adamm Exp adamm $";

/* Keep/remove leading colon as desired, or s/:/?/ */
static char OPTS[] = ":v";
char	*Prog;

double	assertStrtod	(char *);
void	manpage		(void);
void	unhexify	(unsigned char **);
void	usage		(char *);


int main(int argc, char *argv[])
{
    char    message[BUFSIZ];
    double  num1;
    double  num2;
    char *  p;

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	    Prog = &argv[0][0];
    } else {
	    ++Prog;
    }

    /* Deal with --help and --man */
    if (argc > 1) {
	if (
	    (strcmp(argv[1], "--help") == 0)
	    ||
	    (strcmp(argv[1], "--man") == 0)
	) {
	    manpage();
	    exit(0);
	}
    }

    /* Quick sanity check */
    if (argc != 3) {
	(void) sprintf(message, "expected 2 args; got %d.", argc - 1);
	usage(message);
	exit(1);
    }

    /* Convert the numbers */
    num1 = assertStrtod(argv[1]);
    num2 = assertStrtod(argv[2]);

    /* Compare and show the results */
    if (num1 < num2) {
	printf("less\n");
    } else if (num1 == num2) {
	printf("equal\n");
    } else {
	printf("greater\n");
    }

    exit(0);
} /* main */


double assertStrtod(char *str)
{
    char    message[BUFSIZ];
    double  number;
    char *  p;

    number = strtod(str, &p);
    if (strlen(p) > 0) {
	(void) sprintf(message, "not numeric: '%s'", str);
	usage(message);
	exit(1);
    } else {
	return(number);
    }
} /* assertStrtod */


void manpage(void)
{
    unsigned char *lines[] = {
	#include "numcmp.man.hex"
	0
    };

    unhexify(lines);
    return;
} /* manpage */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }

    return;
} /* unhexify */


void usage(char *message)
{
    (void) fprintf(stderr, "%s: %s\n\n", Prog, message);
    (void) fprintf(stderr, "Use \"%s --man\" for more details.\n", Prog);
    return;
} /* Usage */


/*	===== MAN PAGE STARTS HERE =====
NAME
    numcmp -- compare two numbers

SYNOPSIS
    numcmp num1 num2

DESCRIPTION
    numcmp converts both arguments to signed doubles, compares the two,
    and prints "less", "equal", or "greater" if num1 is less than, equal
    to, or greater than num2.

    The shell doesn't grok floating-point numbers, and sometimes it's
    hard to deal with negative numbers; this program makes dealing with
    both of those a lot easier.

EXIT STATUS
    numcmp will exit with one of the following values:

	0    success
	1    error(s)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====		*/


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
