/**
***	$RCSfile: strcmp.c,v $
***	$Revision: 1.13 $
***	$Date: 2010/01/10 17:07:59 $ UTC (aka GMT)
***	$Author: adamm $
**/

/** Copyright notice at the bottom of this file **/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


static char RCSid[] =
    "$Id: strcmp.c,v 1.13 2010/01/10 17:07:59 adamm Exp adamm $";

char	*Prog;

void	manpage		(void);
void	unhexify	(unsigned char **);
void	usage		(char *);


int main(int argc, char *argv[])
{
    int	    c;
    int	    cmp;
    char    message[BUFSIZ];

    /* Make the program name pretty */
    Prog = strrchr(argv[0], '/');
    if (Prog == NULL) {
	    Prog = &argv[0][0];
    } else {
	    ++Prog;
    }

    /* Deal with --help and --man */
    if (argc > 1) {
	if (
	    (strcmp(argv[1], "--help") == 0)
	    ||
	    (strcmp(argv[1], "--man") == 0)
	) {
	    manpage();
	    exit(0);
	}
    }

    /* One quick sanity check */
    if (argc != 3) {
	(void) sprintf(message,
	    "expected 2 strings, got %d.", argc - 1);
	usage(message);
	exit(1);
    }

    cmp = strcmp(argv[1], argv[2]);
    if (cmp < 0) {
	printf("less\n");
    } else if (cmp == 0) {
	printf("equal\n");
    } else {
	printf("greater\n");
    }

    exit(0);
} /* main */


void manpage(void)
{
    unsigned char *lines[] = {
	#include "strcmp.man.hex"
	0
    };

    unhexify(lines);
    return;
} /* manpage */


void unhexify(unsigned char **ptr)
{
    int hex[] = {
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
        0, 10, 11, 12, 13, 14, 15
    };

    int b;
    int c;

    for ( ; *ptr; *(ptr++)) {
        for (b = 0; b < strlen(*ptr); b += 2) {
            c = (hex[*(*ptr + b)] << 4) + hex[*(*ptr + b + 1)];
            putc((int) c, stdout);
        }
    }

    return;
} /* unhexify */


void usage(char *message)
{
    (void) fprintf(stderr, "%s: %s\n\n", Prog, message);
    (void) fprintf(stderr, "Use \"%s --man\" for more details.\n", Prog);
    return;
} /* Usage */


/*	===== MAN PAGE STARTS HERE =====
NAME
    strcmp -- command-line interface to strcmp(3)

SYNOPSIS
    strcmp str1 str2

DESCRIPTION
    strcmp calls strcmp(3) on the two strings and prints "less",
    "equal", or "greater" if str1 is found, respectively, to be
    lexicographically less than, equal to, or greater than str2.

EXIT STATUS
    strcmp will exit with one of the following values:

	0    sucess
	1    error(s)

SEE ALSO
    strcmp(3)

AUTHOR
    Adam Moskowitz <adamm@menlo.com>
===== MAN PAGE STOPS HERE =====		*/


/**
*** Copyright, 2009, Adam Moskowitz. All rights reserved.
***
*** Any redistribution of this software must retain the above copyright
*** notice, this list of conditions, and the following disclaimer.
***
*** Without specific prior written permission from the copyright holder,
*** you may not charge a fee for the redistribution of this software.
***
*** All other redistribution and use is hereby permitted.
***
*** This software is provided "as is" and without any express or implied
*** warranties, including, without limitation, the implied warranties of
*** merchantability and fitness for a particular purpose.
**/
